@@ -28,13 +28,16 @@
 #include <config.h>
 #include "eel-preferences-box.h"
 
+#include <eel/eel-gtk-extensions.h>
 #include <eel/eel-gtk-macros.h>
 #include <eel/eel-string.h>
-#include <gtk/gtkclist.h>
+#include <gtk/gtkcellrenderertext.h>
 #include <gtk/gtkdialog.h>
+#include <gtk/gtkliststore.h>
 #include <gtk/gtknotebook.h>
 #include <gtk/gtkstock.h>
-#include <eel/eel-gtk-extensions.h>
+#include <gtk/gtktreeselection.h>
+#include <gtk/gtktreeview.h>
 #include <libgnome/gnome-i18n.h>
 #include <libgnome/gnome-util.h>
 
@@ -51,16 +54,16 @@ typedef struct
 
 struct EelPreferencesBoxDetails
 {
-	GtkWidget *category_list;
-	GtkWidget *pane_notebook;
-	GList *panes;
-	char *selected_pane;
-	guint select_row_signal_id;
+	GtkTreeView  *category_view;
+	GtkListStore *category_store;
+	GtkWidget    *pane_notebook;
+	GList        *panes;
+	char         *selected_pane;
 };
 
 /* EelPreferencesBoxClass methods */
-static void      eel_preferences_box_class_init   (EelPreferencesBoxClass *preferences_box_class);
-static void      eel_preferences_box_init         (EelPreferencesBox      *preferences_box);
+static void      eel_preferences_box_class_init         (EelPreferencesBoxClass *preferences_box_class);
+static void      eel_preferences_box_init               (EelPreferencesBox      *preferences_box);
 
 /* GObjectClass methods */
 static void      eel_preferences_box_finalize           (GObject                *object);
@@ -75,16 +78,12 @@ static PaneInfo *pane_info_new                          (const char
 static void      pane_info_free                         (PaneInfo               *info);
 
 /* Category list callbacks */
-static void      category_list_select_row_callback      (GtkCList               *clist,
-							 int                     row,
-							 int                     column,
-							 GdkEventButton         *event,
-							 gpointer                user_data);
+static void      category_list_selection_changed        (GtkTreeSelection       *selection,
+							 gpointer                callback_data);
 
 /* Convience functions */
-static int       preferences_box_find_row               (GtkCList               *clist,
-							 char                   *pane_name);
-static void      user_level_changed_callback            (gpointer                callback_data);
+static GtkTreeIter *preferences_box_find_row            (GtkListStore           *model,
+							 const char             *pane_name);
 
 EEL_CLASS_BOILERPLATE (EelPreferencesBox, eel_preferences_box, GTK_TYPE_HBOX)
 
@@ -101,11 +100,6 @@ static void
 eel_preferences_box_init (EelPreferencesBox *preferences_box)
 {
 	preferences_box->details = g_new0 (EelPreferencesBoxDetails, 1);
-
-	eel_preferences_add_callback_while_alive ("user_level",
-						  user_level_changed_callback,
-						  preferences_box,
-						  G_OBJECT (preferences_box));
 }
 
 /*
@@ -136,6 +130,8 @@ eel_preferences_box_finalize (GObject *object)
 		g_list_free (preferences_box->details->panes);
 	}
 
+	g_object_unref (preferences_box->details->category_store);
+
 	g_free (preferences_box->details->selected_pane);
 	g_free (preferences_box->details);
 	
@@ -147,7 +143,7 @@ eel_preferences_box_finalize (GObject *object)
  */
 static void
 preferences_box_select_pane (EelPreferencesBox *preferences_box,
-			     const char *pane_name)
+			     const char        *pane_name)
 {
 	GList *pane_iterator;
 	
@@ -170,8 +166,10 @@ preferences_box_select_pane (EelPreferencesBox *preferences_box,
 				(GTK_NOTEBOOK (preferences_box->details->pane_notebook), 
 				 g_list_position (preferences_box->details->panes, pane_iterator));
 
-			g_free (preferences_box->details->selected_pane);
-			preferences_box->details->selected_pane = g_strdup (pane_name);
+			if (pane_name != preferences_box->details->selected_pane) {
+				g_free (preferences_box->details->selected_pane);
+				preferences_box->details->selected_pane = g_strdup (pane_name);
+			}
 			return;
 		}
 		
@@ -181,74 +179,108 @@ preferences_box_select_pane (EelPreferencesBox *preferences_box,
 	g_warning ("Pane '%s' could not be found.", pane_name);
 }
 
-static int
-preferences_box_find_row (GtkCList *clist, char *pane_name)
+typedef struct {
+	const char  *pane_name;
+	GtkTreeIter *result_iter;
+} foreach_find_t;
+
+static gboolean
+preferences_model_foreach_find (GtkTreeModel *model,
+				GtkTreePath  *path,
+				GtkTreeIter  *iter,
+				gpointer      data)
 {
-	int i;
-	char *pane = NULL;
-	
-	for (i=0; i < GTK_CLIST (clist)->rows; i++) {
-		gtk_clist_get_text (GTK_CLIST (clist), i, 0, &pane);
-		
-		if (eel_str_is_equal (pane, pane_name)) {
-			return i;
-		}
+	gboolean ret;
+	char *pane_name;
+	foreach_find_t *ctx;
+
+	ctx = data;
+	ret = FALSE;
+
+	gtk_tree_model_get (model, iter, CATEGORY_COLUMN, &pane_name, -1);
+
+	if (!strcmp (ctx->pane_name, pane_name)) {
+		ctx->result_iter = gtk_tree_iter_copy (iter);
+		ret = TRUE;
 	}
-	
-	return -1;
+
+	g_free (pane_name);
+
+	return ret;
+}
+
+static GtkTreeIter *
+preferences_box_find_row (GtkListStore *model,
+			  const char   *pane_name)
+{
+	foreach_find_t ctx;
+
+	ctx.pane_name = pane_name;
+	ctx.result_iter = NULL;
+
+	gtk_tree_model_foreach (GTK_TREE_MODEL (model),
+				preferences_model_foreach_find,
+				&ctx);
+
+	return ctx.result_iter;
 }
 
 static void
 preferences_box_category_list_recreate (EelPreferencesBox *preferences_box)
 {
-	GList *iterator;
-	int row = 0;
+	GList *l;
+	PaneInfo *info;
+	GtkTreeIter iter, *selected_iter;
 
 	g_return_if_fail (EEL_IS_PREFERENCES_BOX (preferences_box));
-	g_return_if_fail (GTK_IS_CLIST (preferences_box->details->category_list));
+	g_return_if_fail (GTK_IS_TREE_VIEW (preferences_box->details->category_view));
 
-	/* Block the select_row signal so that the 1st item doesnt get selected.
-	 * Otherwise, we lose the selected_pane.
-	 */
-	g_assert (preferences_box->details->select_row_signal_id != 0);
-	gtk_signal_handler_block (GTK_OBJECT (preferences_box->details->category_list),
-				  preferences_box->details->select_row_signal_id);
+	gtk_list_store_clear (preferences_box->details->category_store);
 
-	gtk_clist_clear (GTK_CLIST (preferences_box->details->category_list));
-
-	for (iterator = preferences_box->details->panes; iterator != NULL; iterator = iterator->next) {
-		PaneInfo *info = iterator->data;
+	selected_iter = NULL;
+	for (l = preferences_box->details->panes; l != NULL; l = l->next) {
+		info = l->data;
 	
 		g_assert (EEL_IS_PREFERENCES_PANE (info->pane_widget));
 
 		if (eel_preferences_pane_get_num_visible_groups (info->pane_widget) > 0) {
-			char *text_array[NUM_CATEGORY_COLUMNS];
-			
-			text_array[CATEGORY_COLUMN] = info->pane_name;
-			gtk_clist_append (GTK_CLIST (preferences_box->details->category_list), text_array);
+
+			gtk_list_store_append (preferences_box->details->category_store, &iter);
+			gtk_list_store_set (preferences_box->details->category_store,
+					    &iter, CATEGORY_COLUMN, info->pane_name, -1);
 
 			if (eel_str_is_equal (info->pane_name, preferences_box->details->selected_pane)) {
-				row = preferences_box_find_row (GTK_CLIST (preferences_box->details->category_list),
-								info->pane_name);
-				
-				if (row == -1) {
-					row = 0;
-				}
+				selected_iter = gtk_tree_iter_copy (&iter);
 			}
+
+			gtk_tree_model_iter_next (
+				GTK_TREE_MODEL (preferences_box->details->category_store), &iter);
 		}
 	}
 	
-	gtk_signal_handler_unblock (GTK_OBJECT (preferences_box->details->category_list),
-				    preferences_box->details->select_row_signal_id);
-	
-	/* You have to do this to get the highlighted row in the clist to change for some reason */
-	gtk_clist_select_row (GTK_CLIST (preferences_box->details->category_list), row, 0);
+	gtk_widget_queue_resize (GTK_WIDGET (preferences_box->details->category_view));
 
-	category_list_select_row_callback (GTK_CLIST (preferences_box->details->category_list),
-					   row,
-					   0,
-					   NULL,
-					   preferences_box);
+	if (selected_iter == NULL) {
+		if (gtk_tree_model_get_iter_root (
+			GTK_TREE_MODEL (preferences_box->details->category_store), &iter)) {
+			selected_iter = gtk_tree_iter_copy (&iter);
+		} else {
+			g_warning ("No preferenced root");
+			selected_iter = NULL;
+		}
+	}
+
+	if (selected_iter) {
+		gtk_tree_selection_select_iter (
+			gtk_tree_view_get_selection (preferences_box->details->category_view),
+			selected_iter);
+		gtk_tree_iter_free (selected_iter);
+	}
+
+	if (preferences_box->details->selected_pane) {
+		preferences_box_select_pane (preferences_box,
+					     preferences_box->details->selected_pane);
+	}
 }
 
 /*
@@ -281,24 +313,26 @@ pane_info_free (PaneInfo *info)
  * Category list callbacks
  */
 static void
-category_list_select_row_callback (GtkCList *clist,
-				   int row,
-				   int column,
-				   GdkEventButton *event,
-				   gpointer callback_data)
+category_list_selection_changed (GtkTreeSelection *selection,
+				 gpointer          callback_data)
 {
-	const char *pane_name = NULL;
+	GtkTreeIter iter;
+	GtkTreeModel *model;
+	char *pane_name = NULL;
 	
 	g_return_if_fail (EEL_IS_PREFERENCES_BOX (callback_data));
 
-	/* The cast here is needed because of the broken gtk_clist api */
-	if (gtk_clist_get_text (clist, row, CATEGORY_COLUMN, (char **) &pane_name) != 1) {
+	if (!gtk_tree_selection_get_selected (selection, &model, &iter)) {
 		return;
 	}
 
+	gtk_tree_model_get (model, &iter, CATEGORY_COLUMN, &pane_name, -1);
+
 	g_return_if_fail (pane_name != NULL);
 
 	preferences_box_select_pane (EEL_PREFERENCES_BOX (callback_data), pane_name);
+
+	g_free (pane_name);
 }
 
 /*
@@ -307,6 +341,7 @@ category_list_select_row_callback (GtkCList *clist,
 GtkWidget*
 eel_preferences_box_new (void)
 {
+	GtkWidget         *tree_view;
 	EelPreferencesBox *preferences_box;
 
 	preferences_box = EEL_PREFERENCES_BOX
@@ -317,26 +352,32 @@ eel_preferences_box_new (void)
  	gtk_box_set_spacing (GTK_BOX (preferences_box), SPACING_BETWEEN_CATEGORIES_AND_PANES);
 
 	/* The category list */
-	preferences_box->details->category_list = gtk_clist_new (NUM_CATEGORY_COLUMNS);
+	preferences_box->details->category_store = gtk_list_store_new (
+		NUM_CATEGORY_COLUMNS, G_TYPE_STRING);
 
-	preferences_box->details->select_row_signal_id =
-		g_signal_connect (preferences_box->details->category_list, 
-				    "select_row",
-				    G_CALLBACK (category_list_select_row_callback),
-				    preferences_box);
+	tree_view = gtk_tree_view_new_with_model (
+		GTK_TREE_MODEL (preferences_box->details->category_store));
 
-	gtk_clist_set_selection_mode (GTK_CLIST (preferences_box->details->category_list), 
-				      GTK_SELECTION_BROWSE);
+	preferences_box->details->category_view = GTK_TREE_VIEW (tree_view);
 
-	gtk_clist_set_column_auto_resize (GTK_CLIST (preferences_box->details->category_list),
-					  CATEGORY_COLUMN,
-					  TRUE);
+	gtk_tree_view_insert_column_with_attributes
+		(preferences_box->details->category_view,
+		 0, "", gtk_cell_renderer_text_new (),
+		 "text", CATEGORY_COLUMN,
+		 NULL);
+	gtk_tree_view_set_headers_visible (preferences_box->details->category_view, FALSE);
+	gtk_tree_view_columns_autosize (preferences_box->details->category_view);
+	gtk_tree_selection_set_mode (gtk_tree_view_get_selection (preferences_box->details->category_view),
+				     GTK_SELECTION_BROWSE);
+
+	g_signal_connect (
+		gtk_tree_view_get_selection (
+			preferences_box->details->category_view),
+		"changed",
+		G_CALLBACK (category_list_selection_changed),
+		preferences_box);
 	
-	gtk_box_pack_start (GTK_BOX (preferences_box),
-			    preferences_box->details->category_list,
-			    FALSE,
-			    TRUE,
-			    0);
+	gtk_box_pack_start (GTK_BOX (preferences_box), tree_view, FALSE, TRUE, 0);
 
 	/* The gtk notebook that the panes go into. */
 	preferences_box->details->pane_notebook = gtk_notebook_new ();
@@ -350,7 +391,7 @@ eel_preferences_box_new (void)
 			    TRUE,
 			    0);
 
-	gtk_widget_show (preferences_box->details->category_list);
+	gtk_widget_show (tree_view);
 	gtk_widget_show (preferences_box->details->pane_notebook);
 
 	return GTK_WIDGET (preferences_box);
@@ -527,14 +568,6 @@ eel_preferences_box_populate (EelPreferencesBox *preferences_box,
 	eel_preferences_box_update (preferences_box);
 }
 
-static void
-user_level_changed_callback (gpointer callback_data)
-{
-	g_return_if_fail (EEL_IS_PREFERENCES_BOX (callback_data));
-
-	eel_preferences_box_update (EEL_PREFERENCES_BOX (callback_data));
-}
-
 GtkWidget *
 eel_preferences_dialog_new (const char *title,
 			    const EelPreferencesPaneDescription *panes)
@@ -546,7 +579,7 @@ eel_preferences_dialog_new (const char *title,
 	g_return_val_if_fail (title != NULL, NULL);
 	
 	dialog = gtk_dialog_new_with_buttons (title, NULL, 0,
-					      GTK_STOCK_OK, GTK_RESPONSE_OK,
+					      GTK_STOCK_CLOSE, GTK_RESPONSE_OK,
 					      NULL);
 
 	/* Setup the dialog */
@@ -556,8 +589,6 @@ eel_preferences_dialog_new (const char *title,
 	
 	gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_OK);
 
-	eel_gtk_window_set_up_close_accelerator (GTK_WINDOW (dialog));
-
 	preference_box = eel_preferences_box_new ();
 
 	vbox = GTK_DIALOG (dialog)->vbox;
@@ -570,12 +601,12 @@ eel_preferences_dialog_new (const char *title,
 			    TRUE,	/* fill */
 			    0);		/* padding */
 
-	gtk_widget_show (preference_box);
-
 	if (panes != NULL) {
 		eel_preferences_dialog_populate (GTK_WINDOW (dialog), panes);
 	}
 	
+	gtk_widget_show (preference_box);
+
 	return dialog;
 }
 
@@ -632,11 +663,11 @@ eel_preferences_box_for_each_pane (const EelPreferencesBox *preferences_box,
 
 void
 eel_preferences_box_rename_pane (EelPreferencesBox *preferences_box,
-				 const char *pane_name,
-				 const char *new_pane_name)
+				 const char        *pane_name,
+				 const char        *new_pane_name)
 {
-	PaneInfo *pane_info;
-	int row;
+	GtkTreeIter *iter;
+	PaneInfo    *pane_info;
 
 	g_return_if_fail (EEL_IS_PREFERENCES_BOX (preferences_box));
 	g_return_if_fail (eel_strlen (pane_name) > 0);
@@ -652,16 +683,20 @@ eel_preferences_box_rename_pane (EelPreferencesBox *preferences_box,
 		return;
 	}
 
-	row = preferences_box_find_row (GTK_CLIST (preferences_box->details->category_list),
-					pane_info->pane_name);
-	g_assert (row != -1);
+	iter = preferences_box_find_row (preferences_box->details->category_store,
+					 pane_name);
+	if (iter == NULL) {
+		g_warning ("No pane called '%s'", pane_info->pane_name);
+		return;
+	}
 
 	g_free (pane_info->pane_name);
 	pane_info->pane_name = g_strdup (new_pane_name);
-	gtk_clist_set_text (GTK_CLIST (preferences_box->details->category_list),
-			    row,
-			    CATEGORY_COLUMN,
-			    pane_info->pane_name);
+
+	gtk_list_store_set (preferences_box->details->category_store,
+			    iter, 0, pane_info->pane_name, -1);
+
+	gtk_tree_iter_free (iter);
 }
 
 char *
